SystemOrganization addCategory: #'Asm-Generator'!

Object subclass: #AsmBlockScope
	instanceVariableNames: 'parent bindings argCount'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Asm-Generator'!

!AsmBlockScope methodsFor: 'accessing' stamp: 'sig 5/6/2007 08:22'!
addName: aName binding: aBinding

	bindings at: aName put: aBinding.
	^ aBinding.! !

!AsmBlockScope methodsFor: 'accessing' stamp: 'sig 5/6/2007 09:41'!
argCount
	^ argCount! !

!AsmBlockScope methodsFor: 'accessing' stamp: 'sig 5/6/2007 15:11'!
bindingOf: aName
	^ bindings at: aName ifAbsent: [
		parent isNil ifFalse: [ parent bindingOf: aName ] ifTrue: [ self error: 'binding for: ''' , aName , ''' not found.']
		]! !

!AsmBlockScope methodsFor: 'accessing' stamp: 'sig 5/6/2007 08:36'!
initialize
	parent := nil.
	bindings := Dictionary new.
	argCount := 0.! !

!AsmBlockScope methodsFor: 'accessing' stamp: 'sig 5/6/2007 07:43'!
isTopLevel
	^ parent == nil.! !

!AsmBlockScope methodsFor: 'accessing' stamp: 'sig 5/6/2007 07:24'!
parent
	^ parent! !

!AsmBlockScope methodsFor: 'accessing' stamp: 'sig 5/6/2007 07:24'!
parent: anObject
	parent := anObject! !

Object subclass: #AsmCallConvention
	instanceVariableNames: 'blockNum result epilog currentBlock blockMapping generator'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Asm-Generator'!

AsmCallConvention subclass: #AsmApiCall
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Asm-Generator'!

!AsmApiCall class methodsFor: 'as yet unclassified' stamp: 'sig 6/7/2007 11:34'!
pragma
	^ #apicall! !

!AsmCallConvention class methodsFor: 'as yet unclassified' stamp: 'sig 6/7/2007 11:34'!
pragma
	"return a symbol for pragma"
	self subclassResponsibility ! !

!AsmCallConvention methodsFor: 'accessing' stamp: 'sig 7/1/2007 03:17'!
blockMapping
	^ blockMapping! !

!AsmCallConvention methodsFor: 'accessing' stamp: 'sig 7/1/2007 22:53'!
currentBlock
	^ currentBlock! !

!AsmCallConvention methodsFor: 'accessing' stamp: 'sig 7/1/2007 22:53'!
currentBlock: anObject
	currentBlock := anObject! !

!AsmCallConvention methodsFor: 'as yet unclassified' stamp: 'sig 5/13/2007 04:35'!
epilogueBlockFor: aMethod
	self subclassResponsibility ! !

!AsmCallConvention methodsFor: 'call support' stamp: 'sig 7/1/2007 03:01'!
generateCallFor: method from: callingConvention caller: aMedCallMethod
	
	self subclassResponsibility ! !

!AsmCallConvention methodsFor: 'as yet unclassified' stamp: 'sig 7/1/2007 21:17'!
generateMethod: method generator: gen
| blk |
	generator := gen.
	method lazyTemps do: [ :temp |  temp realizeWith: gen machine createTemporaryRegister  ].
	"realize arguments"
	method arguments do: [ :arg |  self realizeArgument: arg  ].

	blk := self prologueBlockFor: method.
	blk ifNotNil: [ self result addBlock: blk ].
	epilog := self epilogueBlockFor: method.
	"visit blocks"
	method blocks do: [:block | self visitBlock: block].
	epilog ifNotNil:[ self result addBlock: epilog ].
	"return temps/args back to unrealized state"
	method lazyTemps do: [ :temp |  temp realizeWith:nil ].
	method arguments do: [ :arg |  arg realizeWith: nil ].

	^ self result.

	! !

!AsmCallConvention methodsFor: 'accessing' stamp: 'sig 7/1/2007 20:02'!
generator
	^ generator! !

!AsmCallConvention methodsFor: 'initialize-release' stamp: 'sig 6/27/2007 10:51'!
initialize
	blockNum := 1.
	result := MedMethod new.
	blockMapping := IdentityDictionary new.! !

!AsmCallConvention methodsFor: 'accessing' stamp: 'sig 7/2/2007 01:25'!
machineRegister: aName
	^ generator machine machineRegister: aName! !

!AsmCallConvention methodsFor: 'as yet unclassified' stamp: 'sig 6/7/2007 11:39'!
newBlock
	| blk |
	blk := MedBlock number: blockNum.
	blockNum := blockNum + 1.
	^ blk! !

!AsmCallConvention methodsFor: 'as yet unclassified' stamp: 'sig 5/13/2007 04:35'!
prologueBlockFor: aMethod
	self subclassResponsibility ! !

!AsmCallConvention methodsFor: 'as yet unclassified' stamp: 'sig 5/13/2007 04:18'!
realizeArgument: aScopeArgument
	self subclassResponsibility ! !

!AsmCallConvention methodsFor: 'accessing' stamp: 'sig 5/13/2007 13:13'!
result
	^ result! !

!AsmCallConvention methodsFor: 'translating' stamp: 'sig 6/27/2007 12:51'!
visitArithmatic: aMedArithmatic 
	^ MedArithmatic operator: aMedArithmatic operator arg1: (aMedArithmatic arg1 visitWith: self )
		arg2: (aMedArithmatic arg2 visitWith: self )
! !

!AsmCallConvention methodsFor: 'translating' stamp: 'sig 7/1/2007 23:39'!
visitBlock: aBlock
	| mappedBlock |
	mappedBlock := self blockMapping at: aBlock ifAbsentPut: [self newBlock].
	self result addBlock: mappedBlock.
	self currentBlock: mappedBlock.
	aBlock do: [:expr | | res |
		res := expr visitWith: self.
		res isCollection ifTrue: [ res do: [ :each |  self currentBlock add: each ] ]
		ifFalse: [ (res isNil or: [ res isRegister ]) ifFalse: [self currentBlock add: res] ].
	].

! !

!AsmCallConvention methodsFor: 'translating' stamp: 'sig 7/1/2007 23:13'!
visitCallMethod: aMedCallMethod 
	| callee res |
	callee := aMedCallMethod callee.
	callee isIndirect ifTrue: [
		self notYetImplemented 
		]
	ifFalse: [ "generate call to a method with given arguments, honing current call convention rules and callee rules"
		| convention method |
		method := generator parsedMethod: callee selector ofClass: callee methodClass.
		callee inline ifTrue: [
			"force method to be inlined"
			convention := AsmInlineCall new outerConv: self.
			]
		ifFalse: [ convention := generator conventionByName: method defaultCallConvention. ].
		self assert: [ convention notNil ] description: 'Unsupported call convention for method: ' , callee methodClass name , '>>' , callee selector.
		res := convention generateCallFor: method from: self caller: aMedCallMethod.
		"continue in new block"
		self currentBlock: self newBlock.
		self result addBlock: self currentBlock.
		^ res
	]
	! !

!AsmCallConvention methodsFor: 'translating' stamp: 'sig 6/30/2007 17:29'!
visitComparision: aMedComparision 
	^ MedComparision operator: #cmp arg1: (aMedComparision arg1 visitWith: self)  arg2: (aMedComparision arg2 visitWith: self)! !

!AsmCallConvention methodsFor: 'translating' stamp: 'sig 6/28/2007 08:45'!
visitDivide: aMedDivide 
	^ MedDivide emit: (aMedDivide arg1 visitWith: self) divide: (aMedDivide arg2 visitWith: self ) remander: (aMedDivide remainder visitWith: self)! !

!AsmCallConvention methodsFor: 'translating' stamp: 'sig 7/2/2007 01:36'!
visitJump: aMedJump 
	| mappedBlock |
	self assert: [aMedJump target isBlock].
	mappedBlock := self blockMapping at: aMedJump target ifAbsentPut: [self newBlock].
	^ MedJump type: aMedJump type target: mappedBlock! !

!AsmCallConvention methodsFor: 'translating' stamp: 'sig 6/27/2007 11:04'!
visitLiteral: aLiteral
	^ aLiteral
	! !

!AsmCallConvention methodsFor: 'translating' stamp: 'sig 7/2/2007 01:37'!
visitMachineRegister: aMedMachineRegister 
	^ self machineRegister: aMedMachineRegister name! !

!AsmCallConvention methodsFor: 'translating' stamp: 'sig 6/28/2007 08:41'!
visitMem: aMedMem 
	^ MedMem address: (aMedMem address visitWith: self)! !

!AsmCallConvention methodsFor: 'translating' stamp: 'sig 6/27/2007 11:04'!
visitMov: aMedMov 
	| fromRegister toRegister |
	fromRegister :=(aMedMov from visitWith: self).
	toRegister := (aMedMov to visitWith: self).
	^ MedMov from: fromRegister to: toRegister
! !

!AsmCallConvention methodsFor: 'translating' stamp: 'sig 7/1/2007 02:27'!
visitRet: aMedRet
	^ MedRet new! !

!AsmCallConvention methodsFor: 'translating' stamp: 'sig 7/1/2007 02:26'!
visitReturn: aMedReturn 
	self subclassResponsibility ! !

!AsmCallConvention methodsFor: 'translating' stamp: 'sig 7/2/2007 01:55'!
visitTemporaryRegister: aMedTemporaryRegister 
	^ self generator machine temporaryNamed: aMedTemporaryRegister name! !

AsmCallConvention subclass: #AsmCdeclCall
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Asm-Generator'!

!AsmCdeclCall class methodsFor: 'as yet unclassified' stamp: 'sig 6/7/2007 11:32'!
pragma
	^ #cdecl! !

!AsmCdeclCall methodsFor: 'as yet unclassified' stamp: 'sig 7/2/2007 00:10'!
epilogueBlockFor: aMethod
	| block |
	block := self newBlock.
	block add: (MedMov from: (self machineRegister: #ebp)  to: (self machineRegister: #esp) ).
	block add: (MedPop argument: (self machineRegister: #ebp) ).
	block add: (MedRet new).
	^ block.
! !

!AsmCdeclCall methodsFor: 'as yet unclassified' stamp: 'sig 7/2/2007 00:11'!
prologueBlockFor: aMethod
	| block |
	block := self newBlock.
	block add: (MedPush argument: (self machineRegister: #ebp) ).
	block add: (MedMov from: (self machineRegister: #esp)  to: ( self machineRegister: #ebp) ).
	^ block.! !

!AsmCdeclCall methodsFor: 'as yet unclassified' stamp: 'sig 7/2/2007 01:27'!
realizeArgument: aScopeArgument
	aScopeArgument realizeWith: ( MedMem address: (MedArithmatic operator: #add arg1: (self machineRegister: #ebp) arg2: (MedLiteral literal: -4*(aScopeArgument index) ) ) )! !

!AsmCdeclCall methodsFor: 'as yet unclassified' stamp: 'sig 7/1/2007 02:21'!
visitRet: aRet
	^ MedJump type: #jmp target: epilog! !

!AsmCdeclCall methodsFor: 'as yet unclassified' stamp: 'sig 7/2/2007 01:26'!
visitReturn: aMedReturn 
	^ { 
		MedMov from: (aMedReturn value visitWith: self)  to: (self machineRegister: #eax) .
		MedJump type: #jmp target: epilog
	}

	
	! !

AsmCallConvention subclass: #AsmInlineCall
	instanceVariableNames: 'outerConv arguments returnReg'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Asm-Generator'!

!AsmInlineCall class methodsFor: 'as yet unclassified' stamp: 'sig 6/7/2007 11:32'!
pragma
	^ #inline! !

!AsmInlineCall methodsFor: 'as yet unclassified' stamp: 'sig 7/1/2007 03:19'!
blockMapping
	^ outerConv blockMapping! !

!AsmInlineCall methodsFor: 'as yet unclassified' stamp: 'sig 7/1/2007 22:53'!
currentBlock
	^ outerConv currentBlock! !

!AsmInlineCall methodsFor: 'as yet unclassified' stamp: 'sig 7/1/2007 22:53'!
currentBlock: aBlock
	^ outerConv currentBlock: aBlock! !

!AsmInlineCall methodsFor: 'as yet unclassified' stamp: 'sig 7/1/2007 23:40'!
epilogueBlockFor: aMethod
	^ self newBlock.
! !

!AsmInlineCall methodsFor: 'as yet unclassified' stamp: 'sig 7/1/2007 21:04'!
generateCallFor: method from: callingConvention caller: aMedCallMethod
	"this is entry point, where we start workin with method inlining"
	outerConv := callingConvention.
	returnReg := aMedCallMethod retValueReg.
	arguments := aMedCallMethod args.
	super generateMethod: method generator: outerConv generator.
	^  returnReg visitWith: self

! !

!AsmInlineCall methodsFor: 'as yet unclassified' stamp: 'sig 7/1/2007 02:41'!
generateMethod: method generator: gen
	" i don't expect to be called in such way. inline methods get parsed by calling #generateCallFor:from:arguments:"
	self shouldNotImplement! !

!AsmInlineCall methodsFor: 'as yet unclassified' stamp: 'sig 7/1/2007 02:16'!
newBlock
	^ outerConv newBlock! !

!AsmInlineCall methodsFor: 'as yet unclassified' stamp: 'sig 7/1/2007 02:16'!
outerConv: anOuterCallConvention
	outerConv := anOuterCallConvention.! !

!AsmInlineCall methodsFor: 'as yet unclassified' stamp: 'sig 7/1/2007 20:23'!
prologueBlockFor: aMethod
	^ nil! !

!AsmInlineCall methodsFor: 'as yet unclassified' stamp: 'sig 7/1/2007 20:19'!
realizeArgument: aScopeArgument
	aScopeArgument realizeWith: (arguments at: aScopeArgument index)! !

!AsmInlineCall methodsFor: 'as yet unclassified' stamp: 'sig 7/1/2007 03:19'!
result
	^ outerConv result! !

!AsmInlineCall methodsFor: 'as yet unclassified' stamp: 'sig 7/1/2007 23:36'!
visitReturn: aMedReturn 
	^ { 	MedMov from: (aMedReturn value visitWith: self)  to: (returnReg visitWith: self).
		MedJump type: #jmp target: epilog }

	! !

AsmCallConvention subclass: #AsmPureCall
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Asm-Generator'!

!AsmPureCall commentStamp: 'sig 7/1/2007 02:48' prior: 0!
My instance used for creating a pure assembler routine from AsmMethod instance.
No prologue/epilogue generated for method, method must have no parameters, all returns in method converted to simple ret.!

!AsmPureCall class methodsFor: 'as yet unclassified' stamp: 'sig 6/30/2007 17:48'!
pragma
	^ #pure! !

!AsmPureCall methodsFor: 'as yet unclassified' stamp: 'sig 6/30/2007 19:14'!
generateMethod: method generator: gen
	self assert: [ method arguments isEmpty ] description: 'A pure assembler methods can''t have arguments.'.
	
	^ super generateMethod: method generator: gen! !

Object subclass: #AsmCodeGenerator
	instanceVariableNames: 'machine methods'
	classVariableNames: 'VMFnPointers VMVarPointers'
	poolDictionaries: ''
	category: 'Asm-Generator'!

!AsmCodeGenerator class methodsFor: 'class initialization' stamp: 'sig 7/7/2007 09:43'!
initVMPointers
	VMFnPointers := Dictionary new.
	(ObjectMemory requiredMethodNames union: Interpreter requiredMethodNames) do:  
		[:vname |  | ptr |
			ptr := self primFnPointer: vname.
			ptr ifNotNil: [ VMFnPointers at: vname put: ptr ]
		].

	 VMVarPointers := Dictionary new.
	(Interpreter instVarNames union: ObjectMemory instVarNames) do: [:vname |  | ptr |
		ptr := self primVarPointer: vname.
		ptr ifNotNil: [ VMVarPointers at: vname asSymbol put: ptr ]
		].! !

!AsmCodeGenerator class methodsFor: 'class initialization' stamp: 'sig 7/7/2007 09:43'!
initialize
	self initVMPointers! !

!AsmCodeGenerator class methodsFor: 'primitives' stamp: 'sig 7/7/2007 07:40'!
primFnPointer: aName
<primitive: 'primitiveFnPtr' module: 'VMPointersPlugin'>! !

!AsmCodeGenerator class methodsFor: 'primitives' stamp: 'sig 7/7/2007 05:04'!
primVarPointer: aName
<primitive: 'primitiveGlobalVarPtr' module: 'VMPointersPlugin'>! !

!AsmCodeGenerator class methodsFor: 'class variables' stamp: 'sig 7/7/2007 09:41'!
vmFnPointers
	^ VMFnPointers! !

!AsmCodeGenerator class methodsFor: 'class variables' stamp: 'sig 7/7/2007 09:41'!
vmVarPointers
	^ VMVarPointers! !

!AsmCodeGenerator methodsFor: 'as yet unclassified' stamp: 'sig 7/2/2007 02:09'!
assembleMethod: aMethod
| optimiser result selector withRegisters |
	" low level optimizer"
	result := aMethod.
	optimiser := LowLevelOptimiser from: result machine: machine.
	result := optimiser run.
	selector := InstructionSelector 
		on: result
		for: machine.
	result := selector result.
		
	withRegisters := ColouringRegisterAllocator 
		on: result
		for: machine.
	result := withRegisters result.
	result := JumpRemover 
				removeJumpsFrom: result
				machine: machine.
	^ result! !

!AsmCodeGenerator methodsFor: 'as yet unclassified' stamp: 'sig 7/1/2007 01:46'!
conventionByName: aSymbol
	"return an AsmCallConvention subclass instance or nil, if no convention with given name found"
	AsmCallConvention allSubclassesDo: [ :each |
		each pragma == aSymbol ifTrue: [ ^ each new ]
		].
	^ nil! !

!AsmCodeGenerator methodsFor: 'as yet unclassified' stamp: 'sig 7/2/2007 02:09'!
genAssemblerByteCodes: aMethod
 | convention result asm | 
	machine := MachineX86 new.
	convention := (self conventionByName: aMethod defaultCallConvention) ifNil: [ self error: 'Unsupported call convention: ', convention printString].
	result := convention generateMethod: aMethod generator: self.
	result := self assembleMethod: result.
	asm := AssemblerExupery 
		assemble: result
		withPlugin: machine
		relocator: self.
	^ asm byteCodes.! !

!AsmCodeGenerator methodsFor: 'as yet unclassified' stamp: 'sig 7/2/2007 02:09'!
genAssemblerIntermediate: aMethod
 | convention result  | 
	machine := MachineX86 new.
	convention := (self conventionByName: aMethod defaultCallConvention) ifNil: [ self error: 'Unsupported call convention: ', convention printString].
	result := convention generateMethod: aMethod generator: self.
	^ self assembleMethod: result.
! !

!AsmCodeGenerator methodsFor: 'as yet unclassified' stamp: 'sig 7/4/2007 00:53'!
genAssembly: aMethod
 | convention result  | 
	machine := MachineX86 new.
	convention := (self conventionByName: aMethod defaultCallConvention) ifNil: [ self error: 'Unsupported call convention: ', convention printString].
	result := convention generateMethod: aMethod generator: self.

	^ self assembleMethod: result.
! !

!AsmCodeGenerator methodsFor: 'as yet unclassified' stamp: 'sig 5/9/2007 16:09'!
getMethod: aSelector ofClass: aClass ifAbsent: aBlock
	
	^ (methods at: aClass ifAbsent: [ ^ aBlock value ]) at: aSelector ifAbsent: aBlock.! !

!AsmCodeGenerator methodsFor: 'as yet unclassified' stamp: 'sig 5/9/2007 21:36'!
initialize
	methods := Dictionary new.! !

!AsmCodeGenerator methodsFor: 'accessing' stamp: 'sig 5/13/2007 04:13'!
machine
	^ machine! !

!AsmCodeGenerator methodsFor: 'as yet unclassified' stamp: 'sig 5/9/2007 16:50'!
parseMethod: aMethod ofClass: aClass 
	| method classdict parser |

	method := Parser new parse: (aClass sourceCodeAt: aMethod) string class: aClass.
	method := method asTranslationMethodOfClass: DummyTranslationNode.

	parser := AsmParser generator: self selectorsClass: AsmSelectors.
	method := parser parseMethodFromTranslation: method ofClass: aClass selector: aMethod.

	classdict := methods at: aClass ifAbsent: [ methods at: aClass put: Dictionary new].
	^ classdict at: aMethod put: method.
! !

!AsmCodeGenerator methodsFor: 'as yet unclassified' stamp: 'sig 7/1/2007 01:35'!
parsedMethod: aSelector ofClass: aClass 
	"first, we try to get already parsed method, if not, try parse it"
	^ self getMethod: aSelector ofClass: aClass ifAbsent: [ self parseMethod: aSelector ofClass: aClass  ].
! !

!AsmCodeGenerator methodsFor: 'as yet unclassified' stamp: 'sig 6/27/2007 14:17'!
setUp
	| |
	machine := MachineX86 new.
	methods := Dictionary new.! !

!AsmCodeGenerator methodsFor: 'as yet unclassified' stamp: 'sig 5/6/2007 10:23'!
test
| param1 prolog method eax ebp allocator selector |

	method := MedMethod new.
	prolog := method prolog.

	ebp := machine registerNamed: #ebp.
	eax := machine registerNamed: #eax.
"	param1 := MedMem address: (MedArithmatic operator: #add arg1:  ebp arg2: (MedLiteral literal:-4 ) ).
"	
"	param1 := MedMem address: (MedAddressingMode offset: -4 base: ebp)."

"	prolog add: (MedPush argument: (machine registerNamed: #ebp)).
	prolog add: (MedMov from: (machine registerNamed: #esp) to: ebp ).
"	prolog add: (MedMov from: (MedMov from: eax to: ebp) to: eax ).
"	prolog add: (MedMov from: param1 to: eax ).
"	
	method intermediateForm inspect.
	selector := InstructionSelector on: method for: machine.
	allocator := ColouringRegisterAllocator
	on: selector result
	for: machine.


	^ "allocator result intermediateForm."

	(AssemblerExupery 
		assemble: allocator result
		withPlugin: MockExuperyPlugin
		relocator: MockRelocator new) byteCodes.
! !

Object subclass: #AsmIndirectCallee
	instanceVariableNames: 'expression'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Asm-Generator'!

!AsmIndirectCallee methodsFor: 'as yet unclassified' stamp: 'sig 6/28/2007 18:05'!
expression
	^ expression ! !

!AsmIndirectCallee methodsFor: 'as yet unclassified' stamp: 'sig 6/28/2007 18:05'!
expression: anExpression
	expression := anExpression! !

!AsmIndirectCallee methodsFor: 'as yet unclassified' stamp: 'sig 6/28/2007 18:05'!
intermediateForm
	^ { #Indirect. expression intermediateForm }! !

!AsmIndirectCallee methodsFor: 'as yet unclassified' stamp: 'sig 7/1/2007 01:37'!
isIndirect
	^ true! !

Object subclass: #AsmMethodCallee
	instanceVariableNames: 'selector methodClass inline'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Asm-Generator'!

!AsmMethodCallee methodsFor: 'as yet unclassified' stamp: 'sig 6/28/2007 19:50'!
inline
	^ inline! !

!AsmMethodCallee methodsFor: 'as yet unclassified' stamp: 'sig 6/30/2007 22:21'!
intermediateForm
	^ { #Method. methodClass name.  selector. inline }! !

!AsmMethodCallee methodsFor: 'as yet unclassified' stamp: 'sig 7/1/2007 01:37'!
isIndirect
	^ false! !

!AsmMethodCallee methodsFor: 'as yet unclassified' stamp: 'sig 6/28/2007 18:07'!
methodClass
	^ methodClass! !

!AsmMethodCallee methodsFor: 'as yet unclassified' stamp: 'sig 6/28/2007 18:07'!
selector
	^ selector! !

!AsmMethodCallee methodsFor: 'as yet unclassified' stamp: 'sig 6/28/2007 19:50'!
selector: aSelector methodClass: aMethodClass inline: aBoolean
	selector := aSelector.
	methodClass := aMethodClass.
	inline := aBoolean! !

Object subclass: #AsmParseTests
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Asm-Generator'!

!AsmParseTests class methodsFor: 'as yet unclassified' stamp: 'sig 6/30/2007 22:56'!
calltest1
	| gen method |
	gen := AsmCodeGenerator new.
	method := gen parseMethod: #callmessage ofClass: self.
	^ method.
"	^ gen genAssembly: method."! !

!AsmParseTests class methodsFor: 'as yet unclassified' stamp: 'sig 6/28/2007 16:30'!
calltest1imm
	| gen method |
	gen := AsmCodeGenerator new.
	^ method := gen parseMethod: #callmessage ofClass: self.
! !

!AsmParseTests class methodsFor: 'as yet unclassified' stamp: 'sig 5/9/2007 21:31'!
test1
	^ AsmCodeGenerator new parseMethod: #message1 ofClass: self! !

!AsmParseTests class methodsFor: 'as yet unclassified' stamp: 'sig 6/30/2007 17:16'!
test2
	| gen method |
	gen := AsmCodeGenerator new.
	method := gen parseMethod: #message2 ofClass: self.
	^ gen genAssembly: method.
! !

!AsmParseTests class methodsFor: 'as yet unclassified' stamp: 'sig 6/27/2007 10:42'!
test5
	| gen method |
	gen := AsmCodeGenerator new.
	method := gen parseMethod: #message1 ofClass: self.
	^ gen genAssembly: method.! !

!AsmParseTests class methodsFor: 'as yet unclassified' stamp: 'sig 6/28/2007 08:37'!
test6
	| gen method |
	gen := AsmCodeGenerator new.
	method := gen parseMethod: #message4:with: ofClass: self.
	^ gen genAssembly: method.! !

!AsmParseTests class methodsFor: 'as yet unclassified' stamp: 'sig 7/2/2007 01:06'!
testAlloc
| machine method 
 optimiser result selector withRegisters |

machine := MachineX86 new.
method := MedMethod createIntermediate:
				#(#(#block 1 #(#push #ebp) #(#mov #esp #ebp))
#(#block 3 #(#mov #(#mem #(#add #ebp -4)) t1) #(#mov #(#mem #(#add
#ebp -8)) t2))
#(#block 5 #(#mov #(#add t1 t2) t3) #(#jmp #block4))
#(#block 4)
#(#block 6 #(#mov t3 #eax) #(#jmp #block2))
#(#block 2 #(#mov #ebp #esp) #(#pop #ebp) #(#ret)))
				for: machine.
				
	" low level optimizer"
	result := method.
	optimiser := LowLevelOptimiser from: result machine: machine.
	result := optimiser run.
	selector := InstructionSelector 
		on: result
		for: machine.
	result := selector result.
		
	withRegisters := ColouringRegisterAllocator 
		on: result
		for: machine.
	result := withRegisters result.
	^ result.! !

!AsmParseTests class methodsFor: 'as yet unclassified' stamp: 'sig 7/1/2007 20:01'!
testInline
	| gen method |
	gen := AsmCodeGenerator new.
	method := gen parseMethod: #inlineMessage4:with: ofClass: self.
	^ gen genAssembly: method.! !

!AsmParseTests methodsFor: 'as yet unclassified' stamp: 'sig 6/30/2007 22:54'!
callmessage

" two ways to indicate which method to call: 
	self message ....   - will call the method in same class
	ClassName message .. - will call the method in ClassName class 
	
	in both cases methods must be defined at instance level, not class level!!
"
	self pragma: #cdecl.
	^ self call: (self message4: 1 with: 10)! !

!AsmParseTests methodsFor: 'as yet unclassified' stamp: 'sig 6/30/2007 22:53'!
callmessage: a with: b

" two ways to indicate which method to call: 
	self message ....   - will call the method in same class
	ClassName message .. - will call the method in ClassName class 
	
	in both cases methods must be defined at instance level, not class level!!
"
	self pragma: #cdecl.
	^ self call: (self message4: a with:b)! !

!AsmParseTests methodsFor: 'as yet unclassified' stamp: 'sig 6/30/2007 22:23'!
inlineMessage4: a with: b
	self pragma: #cdecl.
	^ self inlineCall: (self	message4: a with:b)
! !

!AsmParseTests methodsFor: 'as yet unclassified' stamp: 'sig 5/13/2007 13:14'!
message1
	| a b c |
	self pragma: #cdecl.
	c := 1.
	b := 2.
	a := b + c.
	^ a! !

!AsmParseTests methodsFor: 'as yet unclassified' stamp: 'sig 6/30/2007 17:22'!
message2
	| a b c |
	self pragma: #cdecl.
	c := 1.
	b := 2.
	a := b + c.
	a cmp: b. 
	self jumpGreater: #greater.
	self return: c.
	self label: #greater.
	^ a + 1! !

!AsmParseTests methodsFor: 'as yet unclassified' stamp: 'sig 7/4/2007 00:52'!
message4: arg1 with: arg2
	self pragma: #cdecl.
	^ arg1 - arg2.
! !

Object subclass: #AsmParser
	instanceVariableNames: 'topScope scope method currentBlock asmSelectors generator labels noresult'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Asm-Generator'!

!AsmParser class methodsFor: 'as yet unclassified' stamp: 'sig 5/9/2007 08:29'!
generator: aCodeGenerator 
	^ self generator: aCodeGenerator selectorsClass: AsmSelectors! !

!AsmParser class methodsFor: 'as yet unclassified' stamp: 'sig 5/9/2007 08:26'!
generator: aCodeGenerator selectorsClass: aSelectorsClass
	^ self basicNew generator: aCodeGenerator selectorsClass: aSelectorsClass! !

!AsmParser methodsFor: 'parsing processing' stamp: 'sig 5/9/2007 15:19'!
addArguments: anArguments
"called from DummyTranslationNode"
| args count |
	args := OrderedCollection new.
	count := 0.
	
	anArguments do: [ :each | |arg|
		count  := count + 1.
		arg := args add: (MedScopeArgument new name: each name asSymbol index: count).
		scope addName: each name asSymbol binding: arg.
		].
	method setArguments: args.! !

!AsmParser methodsFor: 'parsing processing' stamp: 'sig 6/30/2007 16:25'!
addJump: aJumpType label: aLabelNode
	self assert: [ aLabelNode isConstant and: [aLabelNode value isSymbol] ] description: 'Jump target must be a symbol'.

	labels at: aLabelNode value ifAbsentPut: [ nil ]. "mark label as used"
	^ MedJump type: aJumpType target: aLabelNode value.
! !

!AsmParser methodsFor: 'parsing processing' stamp: 'sig 6/30/2007 16:30'!
addLabel: aSymbol
| blk |
	self assert: [ aSymbol isConstant and: [aSymbol value isSymbol] ] description: 'Label must have symbolic value'.
	"if label already present in dictionary, it's value must be nil"
	blk := labels at: aSymbol value ifAbsent: [nil].
	blk ifNotNil: [ ^ self error: 'Duplicated label: ', aSymbol value printString ].
	currentBlock := method newBlock.
	labels at: aSymbol value put: currentBlock.
	^ self
! !

!AsmParser methodsFor: 'parsing processing' stamp: 'sig 5/9/2007 15:18'!
addLocal: aName
	
	^ scope addName: aName binding: (method newTempNamed: aName asSymbol)! !

!AsmParser methodsFor: 'parsing processing' stamp: 'sig 6/30/2007 16:31'!
emit: aMed
	aMed ifNil: [ ^self ].
	self assert: [ aMed isKindOf: MedLanguage ] description: 'invalid syntax'.
	aMed isRegister ifFalse:[ currentBlock add: aMed ]! !

!AsmParser methodsFor: 'initialize-release' stamp: 'sig 6/30/2007 15:32'!
generator: aCodeGenerator selectorsClass: aSelectorsClass
	generator := aCodeGenerator.
	asmSelectors := aSelectorsClass new parser: self.
	labels := Dictionary new.! !

!AsmParser methodsFor: 'accessing' stamp: 'sig 5/9/2007 22:25'!
method
	^ method! !

!AsmParser methodsFor: 'parsing processing' stamp: 'sig 6/30/2007 17:00'!
oldvisitBlock: aBlock boundArgs: bArgs resultReg: aReg
" Parse a block, using bound arguments, if any, Moving last statement result in aReg, returning MedBlock istance "
	| newblk numArgs lastStmt |
	self flag: #remove.
	self assert: [ aBlock isMemberOf: TStmtListNode ] description: 'Block exprected'.
	numArgs := bArgs isNil ifTrue: [ 0 ] ifFalse: [bArgs size].
	
	self assert: [ aBlock args size = numArgs ] description: 'Block with ', numArgs asString , ' arguments exprected'.
	
	"create new block and scope"
	newblk := method newBlock.
	currentBlock := newblk.
	scope := AsmBlockScope new parent: scope.
	
	numArgs > 0 ifTrue: [
		1 to numArgs do: [:i | scope addName: (aBlock args at: i) name binding: (bArgs at: i)	]
	].
	 
	aBlock isNilStmtListNode ifFalse: [
		aBlock statements allButLastDo: [ :stmt |  self visitStatement: stmt ].
		lastStmt := aBlock statements last.
		"check if last statement is return statement"
		lastStmt isReturn ifFalse: [
				self emit: (MedMov from: (self visitExpression: lastStmt) to: aReg )
			] ifTrue: [
				self visitStatement: lastStmt
			]
		].
	scope := scope parent.
	"currentBlock is the block where we finished parsing and newblk is where we started"
	^ newblk! !

!AsmParser methodsFor: 'parsing processing' stamp: 'sig 6/30/2007 17:00'!
oldvisitJump: aType passBlock: aPassBlock jumpBlock: aJumpBlock

| passBlkEnd reg curr jumpBlk |
self flag: #remove.
	curr := currentBlock.
	reg := method newTemp.
	self visitBlock: aPassBlock boundArgs: nil  resultReg: reg.
	passBlkEnd := currentBlock.
	
	aJumpBlock isNil ifFalse: [
		jumpBlk := self visitBlock: aJumpBlock boundArgs: nil resultReg: reg.		

		" create an unconditional jump from passBlkEnd to new block"
		currentBlock := method newBlock.
		passBlkEnd add: ( MedJump type: #jmp target: currentBlock ).
		
		" add jump to block where we started from"
		curr add: (MedJump type: aType target: jumpBlk ).
		]
	ifTrue: [
		"jumpBlock is nil, simply jump over passBlock to new block"
		currentBlock := method newBlock.
		curr add: (MedJump type: aType target: currentBlock ).
		].
	" return register"
	^ reg.
	! !

!AsmParser methodsFor: 'parsing processing' stamp: 'sig 6/30/2007 17:00'!
oldvisitMessage: msg
| recv |
self flag: #remove.
	recv := msg receiver.
	" check for special case if receiver is self "
	((recv isMemberOf: TVariableNode) and: [ recv name = 'self']) ifTrue: [
		(msg selector beginsWith: #indirectCall:convention:) ifTrue: [
			" indirect call based on expression"
			].		
		msg selector == #inlineCall: ifTrue: [
			^ self visitCall: msg inline: true
			].
		msg selector == #call: ifTrue: [
			^ self visitCall: msg inline: false
			].
		msg selector == #pragma: ifTrue: [
			self assert: [ (msg args at:1) value isSymbol]
				description: 'Argument to pragma must be a symbolic constant'.
			method pragmas add: (msg args at: 1) value.
			^nil
			]
		]
	ifFalse: [
		"for any receiver except self, 
		selector must be one of defined selectors in selectors class"
		^ asmSelectors dispatchOn: msg selector 
			receiver: msg receiver 
			args: msg args.
		]! !

!AsmParser methodsFor: 'parsing processing' stamp: 'sig 5/9/2007 22:25'!
parseInline: aSelector ofClass: aClass arguments: boundArgs
| inlinedMethod oldBlk reg |

	inlinedMethod := generator getMethod: aSelector ofClass: aClass ifAbsent: [
			generator parseMethod: aSelector ofClass: aClass
		].
	oldBlk := currentBlock.
	currentBlock := method newBlock.
	
	reg := method newTemp.
	oldBlk add: (MedInlineMethod new method: inlinedMethod args: boundArgs
		returnPoint: currentBlock returnValue: reg ).

	^ reg.	
! !

!AsmParser methodsFor: 'parsing entry' stamp: 'sig 6/30/2007 17:41'!
parseMethodFromTranslation: aMethodNode ofClass: aClass selector: aSelector

	method := AsmMethod new sourceClass: aClass selector: aSelector.
	currentBlock := method blockNumbered: 1.
	topScope := scope := AsmBlockScope new.
	aMethodNode visitWith: self.
	
	"validate labels"
	labels keysAndValuesDo: [ :key :val |
		self assert: [ val notNil ] description: 'Undefined label: ' , key printString
		].
	"replace jump targets with blocks"
	method blocks do: [ :block | 
		block do: [:each | 
			each isJump ifTrue: [ each initializeType: each type block: (labels at: each target) instruction: nil ]
			]
		].
	^ method! !

!AsmParser methodsFor: 'parsing processing' stamp: 'sig 5/7/2007 06:12'!
visitAndEmitExpression: aExpr
	self emit: (self visitExpression: aExpr)
! !

!AsmParser methodsFor: 'parsing entry' stamp: 'sig 6/30/2007 22:41'!
visitCall: msg inline: aBoolean 
	| regs resultReg methodClass |
	self assert: [ msg isSend ] description: 'A message send must follow call directive'.
	methodClass := (msg receiver isVariable and: [msg receiver name = 'self']) 
					ifTrue: [method sourceClass]
					ifFalse: [Smalltalk at: msg receiver name asSymbol].
	self assert: [ methodClass notNil ] description: 'A receiver of message send must be a class name or ''self'''.
	
	regs := OrderedCollection new.
	msg args do: 
			[:arg | 
			| expr |
			expr := self visitExpression: arg.
			regs add: (expr isRegister 
						ifTrue: [expr]
						ifFalse: 
							[| reg |
							reg := method newTemp.
							self emit: (MedMov from: expr to: reg).
							reg])].
	resultReg := method newTemp.
	self emit: (MedCallMethod new 
				callee: (AsmMethodCallee new 
						selector: msg selector
						methodClass: methodClass
						inline: aBoolean)
				args: regs
				retValueReg: resultReg).
	^ resultReg! !

!AsmParser methodsFor: 'parsing processing' stamp: 'sig 5/7/2007 08:55'!
visitExpression: aNode

	(aNode isMemberOf: TAssignmentNode) ifTrue: [
		| expr |
		expr := self visitExpression: aNode expression.
		self assert: [ expr notNil ] description: 'invalid assignment expression'.
		^ MedMov from: expr to: (scope bindingOf: aNode variable name)
		].
	
	(aNode isMemberOf: TVariableNode) ifTrue: [
		^ scope bindingOf: aNode name
		].
	(aNode isMemberOf: TSendNode) ifTrue: [
		^ self visitMessage: aNode 
		].
	(aNode isMemberOf: TConstantNode) ifTrue: [
		aNode value isInteger ifFalse: [
			^ self error: 'only Integer literals supported'
			].
		^ MedLiteral literal: aNode value
		].
	self error: 'invalid expression'.! !

!AsmParser methodsFor: 'parsing processing' stamp: 'sig 6/30/2007 17:02'!
visitMessage: msg
| recv |
	recv := msg receiver.
	((recv isMemberOf: TVariableNode) and: [ recv name = 'self']) ifTrue: [
		^ asmSelectors dispatchOnDirective: msg selector args: msg args.
		]
	ifFalse: [
		^ asmSelectors dispatchOn: msg selector 
			receiver: msg receiver 
			args: msg args.
		]! !

!AsmParser methodsFor: 'parsing processing' stamp: 'sig 6/30/2007 17:26'!
visitStatement: stmt
	(stmt isMemberOf: TReturnNode) ifTrue: [
		^ self emit: (MedReturn value: (self visitExpression: stmt expression)).
		].
	(stmt isMemberOf: TSendNode) ifTrue: [
		| res |
		res := self visitMessage: stmt.
		(res isKindOf: MedLanguage) ifTrue: [ self emit: res ].
		^ self
		].
	(stmt isMemberOf: TAssignmentNode) ifTrue: [
		^ self visitAndEmitExpression: stmt 
	].

	self error: 'a statement must be one of the following: a method return, message send or assignment '! !

Object subclass: #AsmSelectors
	instanceVariableNames: 'receiver parser'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Asm-Generator'!

!AsmSelectors commentStamp: 'sig 5/7/2007 00:17' prior: 0!
receiver, args is passed without conversion.!

!AsmSelectors methodsFor: 'register-ops' stamp: 'sig 5/7/2007 04:19'!
& aValue
	^ MedArithmatic operator: #and arg1: (parser visitExpression: receiver) arg2: (parser visitExpression: aValue)! !

!AsmSelectors methodsFor: 'register-ops' stamp: 'sig 5/7/2007 04:20'!
* aValue
	^ MedArithmatic operator: #mul arg1: (parser visitExpression: receiver ) arg2: (parser visitExpression: aValue)! !

!AsmSelectors methodsFor: 'register-ops' stamp: 'sig 5/7/2007 04:21'!
+ aValue
	^ MedArithmatic operator: #add arg1: (parser visitExpression: receiver) arg2: (parser visitExpression: aValue)! !

!AsmSelectors methodsFor: 'register-ops' stamp: 'sig 5/7/2007 04:21'!
- aValue
	^ MedArithmatic operator: #sub arg1: (parser visitExpression: receiver) arg2: (parser visitExpression: aValue)! !

!AsmSelectors methodsFor: 'register-ops' stamp: 'sig 5/9/2007 22:26'!
/ aValue
| reg |
	reg := parser method newTemp.
	^ MedDivide emit: (parser visitExpression: receiver) divide: (parser visitExpression: aValue) remander: reg.! !

!AsmSelectors methodsFor: 'register-ops' stamp: 'sig 5/7/2007 04:22'!
<< aValue
	^ MedArithmatic
		operator: #sal
		arg1: (parser visitExpression: receiver)
		arg2: (parser visitExpression: aValue)! !

!AsmSelectors methodsFor: 'register-ops' stamp: 'sig 5/7/2007 05:08'!
= aValue
	^ self cmp: aValue! !

!AsmSelectors methodsFor: 'register-ops' stamp: 'sig 5/7/2007 04:23'!
>> aValue
	^ MedArithmatic
		operator: #sar
		arg1: (parser visitExpression: receiver)
		arg2: (parser visitExpression: aValue)! !

!AsmSelectors methodsFor: 'directive-ops' stamp: 'sig 6/30/2007 22:17'!
call: aMessage
	^ parser visitCall: aMessage inline: false! !

!AsmSelectors methodsFor: 'register-ops' stamp: 'sig 5/7/2007 05:01'!
cmp: aValue
	^ MedComparision 
		operator: #cmp arg1: (parser visitExpression: receiver)  arg2: (parser visitExpression: aValue)! !

!AsmSelectors methodsFor: 'dispatching' stamp: 'sig 6/30/2007 16:02'!
dispatchOn: aSelector receiver: aReceiver args: anArgs
	self assert: [ self isRegisterOp: aSelector] description: 'Unknown selector: ' , aSelector printString.
	receiver := aReceiver.
	^self perform: aSelector withArguments: anArgs
! !

!AsmSelectors methodsFor: 'dispatching' stamp: 'sig 6/30/2007 16:03'!
dispatchOnDirective: aSelector args: anArgs
" directive is message with receiver self"
	self assert: [ self isDirective: aSelector ] description: 'Unknown directive selector: ' , aSelector printString.
	^self perform: aSelector withArguments: anArgs
! !

!AsmSelectors methodsFor: 'register-ops' stamp: 'sig 6/30/2007 16:54'!
div: expression remainder: register
	"divide receiver by expression, put remainder of division into register"
	^ MedDivide emit: (parser visitExpression: receiver) divide: (parser visitExpression: expression) remander: (parser visitExpression: register).! !

!AsmSelectors methodsFor: 'directive-ops' stamp: 'sig 6/30/2007 22:17'!
inlineCall: aMessage
	^ parser visitCall: aMessage inline: true! !

!AsmSelectors methodsFor: 'testing' stamp: 'sig 6/30/2007 15:59'!
isDirective: aSelector
	^ (self class whichCategoryIncludesSelector: aSelector) == #'directive-ops'! !

!AsmSelectors methodsFor: 'testing' stamp: 'sig 6/30/2007 16:00'!
isRegisterOp: aSelector
	^ (self class whichCategoryIncludesSelector: aSelector) == #'register-ops'! !

!AsmSelectors methodsFor: 'directive-ops' stamp: 'sig 6/30/2007 16:26'!
jump: aLabelNode
	^ parser addJump: #jmp label: aLabelNode! !

!AsmSelectors methodsFor: 'directive-ops' stamp: 'sig 6/30/2007 16:27'!
jumpCarry: aLabelNode
	^ parser addJump: #jc label: aLabelNode! !

!AsmSelectors methodsFor: 'directive-ops' stamp: 'sig 6/30/2007 16:27'!
jumpEqual: aLabelNode
	^ parser addJump: #je label: aLabelNode! !

!AsmSelectors methodsFor: 'directive-ops' stamp: 'sig 6/30/2007 16:27'!
jumpGreater: aLabelNode
	^ parser addJump: #jg label: aLabelNode! !

!AsmSelectors methodsFor: 'directive-ops' stamp: 'sig 6/30/2007 16:28'!
jumpGreaterOrEqual: aLabelNode
	^ parser addJump: #jge label: aLabelNode! !

!AsmSelectors methodsFor: 'directive-ops' stamp: 'sig 6/30/2007 16:28'!
jumpLess: aLabelNode
	^ parser addJump: #jl label: aLabelNode! !

!AsmSelectors methodsFor: 'directive-ops' stamp: 'sig 6/30/2007 16:28'!
jumpLessOrEqual: aLabelNode
	^ parser addJump: #jle label: aLabelNode! !

!AsmSelectors methodsFor: 'directive-ops' stamp: 'sig 6/30/2007 16:28'!
jumpNotCarry: aLabelNode
	^ parser addJump: #jnc label: aLabelNode! !

!AsmSelectors methodsFor: 'directive-ops' stamp: 'sig 6/30/2007 16:28'!
jumpNotEqual: aLabelNode
	^ parser addJump: #jne label: aLabelNode! !

!AsmSelectors methodsFor: 'directive-ops' stamp: 'sig 6/30/2007 16:28'!
jumpNotZero: aLabelNode
	^ parser addJump: #jnz label: aLabelNode! !

!AsmSelectors methodsFor: 'directive-ops' stamp: 'sig 6/30/2007 16:28'!
jumpOverflow: aLabelNode
	^ parser addJump: #jo label: aLabelNode! !

!AsmSelectors methodsFor: 'directive-ops' stamp: 'sig 6/30/2007 16:28'!
jumpZero: aLabelNode
	^ parser addJump: #jz label: aLabelNode! !

!AsmSelectors methodsFor: 'directive-ops' stamp: 'sig 6/30/2007 17:17'!
label: aSymbol
	parser addLabel: aSymbol! !

!AsmSelectors methodsFor: 'register-ops' stamp: 'sig 5/7/2007 04:25'!
loadDword
	^ MedMem address: ( parser visitExpression: receiver )! !

!AsmSelectors methodsFor: 'register-ops' stamp: 'sig 5/6/2007 14:13'!
not
" not yet supported "
	self notYetImplemented ! !

!AsmSelectors methodsFor: 'initialization' stamp: 'sig 5/7/2007 04:12'!
parser: aParser
	parser := aParser.! !

!AsmSelectors methodsFor: 'directive-ops' stamp: 'sig 6/30/2007 16:35'!
pop
	| temp |
	temp := parser method newTemp.
	parser emit: (MedPop argument: temp).
	^ temp.! !

!AsmSelectors methodsFor: 'directive-ops' stamp: 'sig 6/30/2007 16:42'!
pragma: aSymbol
	self assert: [ aSymbol isConstant and: [aSymbol value isSymbol] ]
		description: 'Argument of pragma must be a symbolic constant'.
	parser method pragmas add: aSymbol value.! !

!AsmSelectors methodsFor: 'directive-ops' stamp: 'sig 6/30/2007 16:37'!
push: expression
	| expr |
	expr := parser visitExpression: expression.
	self assert: [ expr isKindOf: MedLanguage] description: 'invalid syntax'.
	parser emit: (MedPush argument: expr).
! !

!AsmSelectors methodsFor: 'directive-ops' stamp: 'sig 6/30/2007 16:38'!
return
	^ MedRet new ! !

!AsmSelectors methodsFor: 'directive-ops' stamp: 'sig 6/30/2007 17:15'!
return: anExpression
	^  MedReturn value: (parser visitExpression: anExpression)! !

!AsmSelectors methodsFor: 'register-ops' stamp: 'sig 5/7/2007 04:26'!
storeDword: aValue
	^ MedMov from: (parser visitExpression: aValue) to: (MedMem address: (parser visitExpression: receiver)  )! !

!AsmSelectors methodsFor: 'register-ops' stamp: 'sig 6/30/2007 15:58'!
| aValue
	^ MedArithmatic operator: #or arg1: (parser visitExpression: receiver) arg2: (parser visitExpression: aValue)! !

Object subclass: #DummyTranslationNode
	instanceVariableNames: 'arguments locals block'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Asm-Generator'!

!DummyTranslationNode commentStamp: 'sig 5/10/2007 10:27' prior: 0!
My instances are dummy translation nodes used to start parsing smalltalk method parse tree 
translated to VMMaker's form and then used by AsmParser to generate an intermediate.!

!DummyTranslationNode methodsFor: 'as yet unclassified' stamp: 'sig 5/9/2007 07:57'!
comment: anUndefinedObject 
	"keep this method. get called from MethodNode"! !

!DummyTranslationNode methodsFor: 'as yet unclassified' stamp: 'sig 5/9/2007 08:05'!
setSelector: aSelector args: aArgs locals: aLocals block: aBlockNode primitive: aPrimNumber

	aPrimNumber = 0 ifFalse: [ self error: 'No primitive calls allowed.' ].
	arguments := aArgs.
	locals := aLocals.
	block := aBlockNode asTranslatorNode.
	
! !

!DummyTranslationNode methodsFor: 'as yet unclassified' stamp: 'sig 5/9/2007 07:59'!
visitWith: aParser

	aParser addArguments: arguments.
	locals do: [ :each | aParser addLocal: each name ].
	block statements do: [:each | aParser visitStatement: each]
! !

Object subclass: #VMPointers
	instanceVariableNames: 'zzzVmGlobalPtrs zzzVmFunctionPtrs'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Asm-Generator'!

!VMPointers commentStamp: 'sig 7/7/2007 04:41' prior: 0!
This class is included into VM interp.c
see Interpreter>>generateInterpreterFile!

!VMPointers class methodsFor: 'as yet unclassified' stamp: 'sig 7/7/2007 07:20'!
declareCVarsIn: aCCodeGenerator
	| str |
	str := String streamContents: [:s | | block |
		
		s nextPutAll: 'char * zzzVmGlobalPtrs [] = {'.
		block := [:var |
			(#(child field parentField youngStartLocal currentBytecode localIP localSP localHomeContext localReturnContext localReturnValue) includes: var asSymbol )
			ifFalse: [
				s cr; tab;
				nextPutAll: '"'; nextPutAll: var; nextPutAll: '" , (char*)';
				nextPut: $&;
				nextPutAll: var;
				nextPut: $,
				]
			].
		
		ObjectMemory instVarNames do: block.
		Interpreter instVarNames do: block.
		s nextPutAll: ' 0 }'.
	 ].
	aCCodeGenerator
		var: #zzzVmGlobalPtrs
		declareC: str.

	str := String streamContents: [:s | 
		s nextPutAll: 'char * zzzVmFunctionPtrs [] = {'.
		(ObjectMemory requiredMethodNames union: Interpreter requiredMethodNames) do: [:fnname | 
			s cr; tab;
			nextPutAll: '"'; nextPutAll: fnname; nextPutAll: '" ,(char*)(void*)';
			nextPutAll: (aCCodeGenerator cFunctionNameFor: fnname);
			nextPut: $,
			].
		s nextPutAll: ' 0 }'.
		].		
	aCCodeGenerator
		var: #zzzVmFunctionPtrs
		declareC: str.

	! !

!VMPointers class methodsFor: 'as yet unclassified' stamp: 'sig 7/7/2007 04:21'!
hasHeaderFile
	^ false! !

!VMPointers class methodsFor: 'as yet unclassified' stamp: 'sig 7/6/2007 22:59'!
requiredMethodNames
	^ #()! !

!VMPointers class methodsFor: 'as yet unclassified' stamp: 'sig 7/7/2007 04:21'!
requiresPlatformFiles
	^ false! !

!VMPointers methodsFor: 'as yet unclassified' stamp: 'sig 7/7/2007 07:45'!
fnPointer: aName
	| ptr namestr namestrlen i c |
	self var: #namestr type: 'char *'.
	self export: true.
	namestr := self firstIndexableField: aName.
	namestrlen := self byteSizeOf: aName.
	ptr := 0.
	[(zzzVmFunctionPtrs at: ptr) = 0] whileFalse: [
		i := 0.
		[ i <namestrlen ] whileTrue: [ 
			c := (zzzVmFunctionPtrs at: ptr) at:i.
			(c = 0 or: [ c ~= (namestr at: i) ]) ifTrue: [ i := namestrlen ].
			i:=i+1 ].
		(i > namestrlen) ifFalse:[ 
			^ self cCode: 'positive32BitIntegerFor((sqInt)zzzVmFunctionPtrs[ptr+1])'.
			].
		ptr := ptr + 2.
		].
	^ self nilObject! !

!VMPointers methodsFor: 'as yet unclassified' stamp: 'sig 7/7/2007 06:55'!
varPointer: aName
	| ptr namestr namestrlen i c |
	self var: #namestr type: 'char *'.
	self export: true.
	namestr := self firstIndexableField: aName.
	namestrlen := self byteSizeOf: aName.
	ptr := 0.
	[(zzzVmGlobalPtrs at: ptr) = 0] whileFalse: [
		i := 0.
		[ i <namestrlen ] whileTrue: [ 
			c := (zzzVmGlobalPtrs at: ptr) at:i.
			(c = 0 or: [ c ~= (namestr at: i) ]) ifTrue: [ i := namestrlen ].
			i:=i+1 ].
		(i > namestrlen) ifFalse:[ 
			^ self cCode: 'positive32BitIntegerFor((sqInt)zzzVmGlobalPtrs[ptr+1])'.
			].
		ptr := ptr + 2.
		].
	^ self nilObject! !

!VMPointers methodsFor: 'as yet unclassified' stamp: 'sig 7/7/2007 04:27'!
varPointers

	^ self cCode: 'positive32BitIntegerFor((sqInt)&zzzVmGlobalPtrs)' inSmalltalk: [ zzzVmGlobalPtrs ]! !

MedTemporaryRegister subclass: #MedLazyTemporary
	instanceVariableNames: 'tempReg'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Asm-Generator'!

!MedLazyTemporary methodsFor: 'as yet unclassified' stamp: 'sig 6/30/2007 22:53'!
intermediateForm
	^ name isNil ifTrue: [
		 '<temp' , (self hash printStringBase: 16) , '>'.
		] ifFalse: [
			name
		]! !

!MedLazyTemporary methodsFor: 'as yet unclassified' stamp: 'sig 6/30/2007 22:51'!
printOn: aStream
	aStream nextPutAll: self intermediateForm! !

!MedLazyTemporary methodsFor: 'as yet unclassified' stamp: 'sig 5/9/2007 09:58'!
realizeWith: aRegister
	tempReg := aRegister.! !

!MedLazyTemporary methodsFor: 'as yet unclassified' stamp: 'sig 5/12/2007 11:30'!
visitWith: anObject
	self assert: [tempReg notNil ] description: 'A lazy temporary register must be realized before it can be used by compiler'.
	^ tempReg visitWith: anObject! !

InterpreterPlugin subclass: #VMPointersPlugin
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Asm-Generator'!

!VMPointersPlugin methodsFor: 'as yet unclassified' stamp: 'sig 7/7/2007 07:39'!
primitiveFnPtr
	| varName |
	
	self export: true.
	varName := interpreterProxy stackValue: 0.
	(interpreterProxy isBytes: varName)
		ifFalse: [^interpreterProxy primitiveFail].
	^ interpreterProxy pop: 1 thenPush: (self fnPointer: varName)! !

!VMPointersPlugin methodsFor: 'as yet unclassified' stamp: 'sig 7/7/2007 05:34'!
primitiveGlobalVarPtr
	| varName |
	
	self export: true.
	varName := interpreterProxy stackValue: 0.
	(interpreterProxy isBytes: varName)
		ifFalse: [^interpreterProxy primitiveFail].
	^ interpreterProxy pop: 1 thenPush: (self varPointer: varName)! !

MedLanguage subclass: #MedCallMethod
	instanceVariableNames: 'callee args retValueReg'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Asm-Generator'!

!MedCallMethod methodsFor: 'as yet unclassified' stamp: 'sig 6/28/2007 17:32'!
args
	^ args! !

!MedCallMethod methodsFor: 'as yet unclassified' stamp: 'sig 6/28/2007 17:31'!
callee
	^ callee! !

!MedCallMethod methodsFor: 'as yet unclassified' stamp: 'sig 6/28/2007 19:55'!
callee: aCallee args: anArgs retValueReg: aRetValueReg 
	callee := aCallee.
	args := anArgs.
	retValueReg := aRetValueReg.
! !

!MedCallMethod methodsFor: 'as yet unclassified' stamp: 'sig 6/28/2007 18:02'!
convention
	^ convention! !

!MedCallMethod methodsFor: 'as yet unclassified' stamp: 'sig 6/30/2007 22:47'!
intermediateForm
	^ {#Call:. (callee intermediateForm).  args asArray. retValueReg intermediateForm. } ! !

!MedCallMethod methodsFor: 'as yet unclassified' stamp: 'sig 6/28/2007 17:32'!
retValueReg
	^ retValueReg! !

!MedCallMethod methodsFor: 'as yet unclassified' stamp: 'sig 6/28/2007 17:31'!
visitWith: aGenerator
	^ aGenerator visitCallMethod: self! !

MedLanguage subclass: #MedScopeArgument
	instanceVariableNames: 'name index replacement'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Asm-Generator'!

!MedScopeArgument methodsFor: 'as yet unclassified' stamp: 'sig 5/6/2007 08:37'!
index
	^ index! !

!MedScopeArgument methodsFor: 'as yet unclassified' stamp: 'sig 5/6/2007 07:15'!
intermediateForm
	^ {#argument. name}! !

!MedScopeArgument methodsFor: 'as yet unclassified' stamp: 'sig 5/6/2007 07:50'!
name: aName
	name := aName! !

!MedScopeArgument methodsFor: 'as yet unclassified' stamp: 'sig 5/6/2007 08:37'!
name: aName index: aIndex
	name := aName.
	index := aIndex! !

!MedScopeArgument methodsFor: 'as yet unclassified' stamp: 'sig 5/10/2007 09:03'!
realizeWith: areplacement
	replacement := areplacement.! !

!MedScopeArgument methodsFor: 'as yet unclassified' stamp: 'sig 5/12/2007 11:30'!
visitWith: aVisitor
	self assert: [ replacement notNil ] description: 'MedScopeArgument must have a replacement set'.
	^replacement visitWith: aVisitor! !

MedMethod subclass: #AsmMethod
	instanceVariableNames: 'selector sourceClass arguments lazyTemps pragmas'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Asm-Generator'!

!AsmMethod commentStamp: 'sig 5/9/2007 16:57' prior: 0!
<selector> - a message selector designating the compiled method
<sourceClass> - a class name where method source located
<arguments> - an ordered collection of argument accessors
<lazyTemps> - an ordered collection of lazy temporaries used by method
<pragmas> - a collection of pragmas defined for method!

!AsmMethod class methodsFor: 'as yet unclassified' stamp: 'sig 7/10/2007 00:44'!
setSelector: a args: b locals: c block: d primitive: e! !

!AsmMethod methodsFor: 'accessing' stamp: 'sig 5/9/2007 09:05'!
arguments
	^ arguments! !

!AsmMethod methodsFor: 'accessing' stamp: 'sig 6/27/2007 10:40'!
blocks
	^ blocks! !

!AsmMethod methodsFor: 'accessing' stamp: 'sig 6/30/2007 17:49'!
defaultCallConvention
	pragmas do: [ :pragma |
		(#( #inline #cdecl #apical) includes: pragma) ifTrue: [^ pragma]
		].
	^ #pure.! !

!AsmMethod methodsFor: 'creation' stamp: 'sig 5/9/2007 21:27'!
initialize
	super initialize.
	lazyTemps := OrderedCollection new.
	pragmas := OrderedCollection new.! !

!AsmMethod methodsFor: 'accessing' stamp: 'sig 5/9/2007 15:22'!
lazyTemps
	^ lazyTemps! !

!AsmMethod methodsFor: 'parsing' stamp: 'sig 5/9/2007 12:02'!
newTemp
	^ lazyTemps add: MedLazyTemporary new.
! !

!AsmMethod methodsFor: 'parsing' stamp: 'sig 5/9/2007 12:13'!
newTempNamed: aName
	^ lazyTemps add: (MedLazyTemporary new initializeName: aName asSymbol).
! !

!AsmMethod methodsFor: 'accessing' stamp: 'sig 5/12/2007 12:04'!
pragmas 
	^ pragmas! !

!AsmMethod methodsFor: 'accessing' stamp: 'sig 5/6/2007 05:05'!
selector
	^ selector! !

!AsmMethod methodsFor: 'accessing' stamp: 'sig 5/9/2007 12:15'!
setArguments: anOrderedCollection

" called from parser"
	arguments := anOrderedCollection.! !

!AsmMethod methodsFor: 'accessing' stamp: 'sig 5/9/2007 08:06'!
sourceClass
	^ sourceClass! !

!AsmMethod methodsFor: 'accessing' stamp: 'sig 5/9/2007 16:55'!
sourceClass: anObject selector: aSelector
	sourceClass := anObject.
	selector := aSelector.! !

!VMMaker methodsFor: '*Asm-Generator-override' stamp: 'sig 7/7/2007 04:40'!
generateInterpreterFile
	"Translate the Smalltalk description of the virtual machine into C.  If 'self doInlining' is true, small method bodies are inlined to reduce procedure call overhead.  On the PPC, this results in a factor of three speedup with only 30% increase in code size.  Subclasses can use specialised versions of CCodeGenerator and interpreterClass."

	self needsToRegenerateInterpreterFile ifFalse: [^nil].
	self interpreterClass initialize.
	ObjectMemory initializeWithBytesToWord: self bytesPerWord.
	self createCodeGenerator
		addClass: self interpreterClass;
		addClass: ObjectMemory;
		addClass: VMPointers;
		storeHeaderOnFile: self interpreterHeaderPath bytesPerWord: self bytesPerWord;
		storeCodeOnFile: self interpreterFilePath doInlining: self doInlining! !
AsmCodeGenerator initialize!
